import abc
import typing as ta
import weakref

from .impls import find_impl as default_find_impl


T = ta.TypeVar('T')
U = ta.TypeVar('U')
U_co = ta.TypeVar('U_co', covariant=True)


##


_build_weak_dispatch_cache_impl: ta.Callable | None = None
_build_strong_dispatch_cache_impl: ta.Callable | None = None

try:
    from . import _dispatch  # type: ignore  # noqa
except ImportError:
    pass
else:
    _build_strong_dispatch_cache_impl = _dispatch.build_strong_dispatch_cache


##


class Dispatcher(ta.Generic[T]):
    """Shared dispatching system for functions and methods. Logic directly mimics `functools.singledispatch`."""

    def __init__(
            self,
            find_impl: ta.Callable[[type, ta.Mapping[type, T]], T | None] | None = None,
            *,
            strong_cache: bool = False,
            uncached_miss: bool = False,
    ) -> None:
        super().__init__()

        if find_impl is None:
            find_impl = default_find_impl
        self._find_impl = find_impl
        self._strong_cache = strong_cache
        self._uncached_miss = uncached_miss

        self._impls_by_arg_cls: dict[type, T] = {}

        if strong_cache:
            if _build_strong_dispatch_cache_impl is not None:
                self._cache_factory = _build_strong_dispatch_cache_impl
            else:
                self._cache_factory = Dispatcher._StrongCache  # noqa
        elif _build_weak_dispatch_cache_impl is not None:
            self._cache_factory = _build_weak_dispatch_cache_impl
        else:
            self._cache_factory = Dispatcher._WeakCache  # noqa

        self._reset_cache(None)

    #

    class _CacheParams(ta.NamedTuple):
        impls_by_arg_cls: dict[type, ta.Any]
        find_impl: ta.Callable[[type, ta.Mapping[type, ta.Any]], ta.Any | None]
        reset_cache_for_token: ta.Callable[['Dispatcher._Cache[ta.Any]'], None]
        uncached_miss: bool
        token: ta.Any | None

    class _Cache(ta.Protocol[U_co]):
        @property
        def token(self) -> ta.Any | None: ...

        @property
        def dct(self) -> ta.Mapping[ta.Any, U_co]: ...

        def dispatch(self, cls: type) -> ta.Any | None: ...

    _cache_factory: ta.Callable[[_CacheParams], _Cache[T]]

    _cache: _Cache[T]

    #

    class _StrongCache:
        def __init__(self, params: 'Dispatcher._CacheParams') -> None:
            token = params.token
            impls_by_arg_cls = params.impls_by_arg_cls
            find_impl = params.find_impl
            uncached_miss = params.uncached_miss

            dct: dict[type, ta.Any] = {}

            def dispatch(cls: type) -> ta.Any | None:
                if token is not None and abc.get_cache_token() != token:
                    params.reset_cache_for_token(self)
                    return find_impl(cls, impls_by_arg_cls)

                try:
                    return dct[cls]
                except KeyError:
                    pass

                impl: ta.Any | None
                try:
                    impl = impls_by_arg_cls[cls]
                except KeyError:
                    impl = find_impl(cls, impls_by_arg_cls)

                if impl is not None or not uncached_miss:
                    dct[cls] = impl
                return impl

            self.token = token
            self.dct = dct
            self.dispatch: ta.Callable[[type], ta.Any | None] = dispatch

    class _WeakCache:
        def __init__(self, params: 'Dispatcher._CacheParams') -> None:
            token = params.token
            impls_by_arg_cls = params.impls_by_arg_cls
            find_impl = params.find_impl
            uncached_miss = params.uncached_miss

            dct: dict[weakref.ref[type], ta.Any] = {}

            weakref_ref_ = weakref.ref

            def dct_remove(k, self_ref=weakref.ref(self)):  # noqa
                if (ref_self := self_ref()) is not None:
                    try:
                        del ref_self.dct[k]
                    except KeyError:
                        pass

            def dispatch(cls: type) -> ta.Any | None:
                if token is not None and abc.get_cache_token() != token:
                    params.reset_cache_for_token(self)
                    return find_impl(cls, impls_by_arg_cls)

                cls_ref = weakref_ref_(cls)
                try:
                    return dct[cls_ref]
                except KeyError:
                    pass
                del cls_ref

                impl: ta.Any | None
                try:
                    impl = impls_by_arg_cls[cls]
                except KeyError:
                    impl = find_impl(cls, impls_by_arg_cls)

                if impl is not None or not uncached_miss:
                    dct[weakref_ref_(cls, dct_remove)] = impl
                return impl

            self.token = token
            self.dct = dct
            self.dispatch: ta.Callable[[type], ta.Any | None] = dispatch

    #

    def _reset_cache(self, token: ta.Any | None) -> None:
        self._cache = self._cache_factory(Dispatcher._CacheParams(  # noqa
            impls_by_arg_cls=self._impls_by_arg_cls,
            find_impl=self._find_impl,
            reset_cache_for_token=self._reset_cache_for_token,
            uncached_miss=self._uncached_miss,
            token=token,
        ))

    def _reset_cache_for_token(self, prev: _Cache) -> None:
        if prev is None or self._cache is prev:
            self._reset_cache(abc.get_cache_token())

    def dispatch(self, cls: type) -> ta.Any | None:
        return self._cache.dispatch(cls)

    def cache_size(self) -> int:
        return len(self._cache.dct)

    def register(self, impl: T, cls_col: ta.Iterable[type]) -> T:
        has_token = self._cache.token is not None

        for cls in cls_col:
            self._impls_by_arg_cls[cls] = impl

            if not has_token and hasattr(cls, '__abstractmethods__'):
                has_token = True

        self._reset_cache(abc.get_cache_token() if has_token else None)
        return impl



"""
TODO:
 - @lang.copy_type
"""
import typing as ta

from ... import cached
from ... import check
from ... import dataclasses as dc
from ... import lang
from ..api.naming import Naming
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory


T = ta.TypeVar('T')


##


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class FieldOptions(lang.Final):
    """
    Unified field options - all configuration for a single field's marshaling/unmarshaling.

    This is the single source of truth for field configuration, merging what used to be split between FieldOptions and
    FieldMetadata in the old design.
    """

    ##
    # Naming

    name: str | None = None
    alts: ta.Iterable[str] | None = None

    ##
    # Behavior options

    omit_if: ta.Callable[[ta.Any], bool] | None = None

    default: lang.Maybe[ta.Any] | None = None

    embed: bool | None = None

    generic_replace: bool | None = None

    no_marshal: bool | None = None
    no_unmarshal: bool | None = None

    ##
    # Custom handlers

    marshaler: Marshaler | None = None
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = None
    unmarshaler_factory: UnmarshalerFactory | None = None

    ##
    # Validation

    def __post_init__(self) -> None:
        check.isinstance(self.default, (lang.Maybe, None))
        check.isinstance(self.marshaler, (Marshaler, None))
        check.isinstance(self.unmarshaler, (Unmarshaler, None))

    ##
    # Merging

    def merge(self, *overrides: ta.Optional['FieldOptions']) -> 'FieldOptions':
        kw: dict[str, ta.Any] = {}
        for obj in [self, *overrides]:
            if obj is None:
                continue

            for fld in _field_options_fields():
                if (fv := getattr(obj, fld.name)) is None:
                    continue

                kw[fld.name] = fv

        if not kw:
            return self
        return FieldOptions(**kw)


@lang.cached_function
def _field_options_fields() -> ta.Sequence[dc.Field]:
    return dc.fields(FieldOptions)


DEFAULT_FIELD_OPTIONS = FieldOptions()


##


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class ObjectSpecials(lang.Final):
    """Special field names for an object."""

    unknown: str | None = None
    source: str | None = None

    @cached.property
    def set(self) -> frozenset[str]:
        return frozenset(v for v in dc.asdict(self).values() if v is not None)


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class ObjectOptions(lang.Final):
    """Object-level marshaling options."""

    ##
    # Naming

    field_naming: Naming | None = None

    ##
    # Behavior options

    ignore_unknown: bool = False

    ##
    # Special fields

    unknown_field: str | None = None
    source_field: str | None = None

    @cached.property
    def specials(self) -> ObjectSpecials:
        return ObjectSpecials(
            unknown=self.unknown_field,
            source=self.source_field,
        )

    ##
    # Field defaults

    field_defaults: FieldOptions = DEFAULT_FIELD_OPTIONS

    ##
    # Merging

    def merge(self, *overrides: ta.Optional['ObjectOptions']) -> 'ObjectOptions':
        kw: dict[str, ta.Any] = {}
        for obj in [self, *overrides]:
            if obj is None:
                continue

            for fld in _object_options_fields():
                if (fv := getattr(obj, fld.name)) is None:
                    continue

                if fld.type is FieldOptions:
                    fv = kw.get(fld.name, DEFAULT_FIELD_OPTIONS).merge(fv)

                kw[fld.name] = fv

        if not kw:
            return self
        return ObjectOptions(**kw)


@lang.cached_function
def _object_options_fields() -> ta.Sequence[dc.Field]:
    return dc.fields(ObjectOptions)


DEFAULT_OBJECT_OPTIONS = ObjectOptions()



"""
https://www.jsonrpc.org/specification

TODO:
 - drop NotSpecified, use lang.Maybe, make marshal do that
 - server/client impl

See:
 - https://github.com/python-lsp/python-lsp-jsonrpc
"""
import operator
import types
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh


T = ta.TypeVar('T')

NUMBER_TYPES: tuple[type, ...] = (int, float)
Number: ta.TypeAlias = int | float

Object: ta.TypeAlias = ta.Mapping[str, ta.Any]

ID_TYPES: tuple[type, ...] = (str, *NUMBER_TYPES, types.NoneType)
Id: ta.TypeAlias = str | Number | None


##


VERSION = '2.0'


##


class NotSpecified(lang.Marker):
    pass


def is_not_specified(v: ta.Any) -> bool:
    return v is NotSpecified


def check_not_not_specified(v: T | type[NotSpecified]) -> T:
    check.arg(not is_not_specified(v))
    return ta.cast(T, v)


##


@dc.dataclass(frozen=True)
@msh.update_fields_options(['id'], omit_if=is_not_specified, default=lang.just(NotSpecified))
@msh.update_fields_options(['params'], omit_if=operator.not_)
class Request(lang.Final):
    id: Id | type[NotSpecified]

    @property
    def is_notification(self) -> bool:
        return self.id is NotSpecified

    def id_value(self) -> Id:
        return check.isinstance(self.id, ID_TYPES)

    method: str
    params: Object | None = None

    jsonrpc: str = dc.field(default=VERSION, kw_only=True)
    dc.validate(lambda self: self.jsonrpc == VERSION)



def request(id: Id, method: str, params: Object | None = None) -> Request:  # noqa
    return Request(id, method, params)


def notification(method: str, params: Object | None = None) -> Request:
    return Request(NotSpecified, method, params)


##


@dc.dataclass(frozen=True)
@msh.update_fields_options(['result', 'error'], omit_if=is_not_specified)
class Response(lang.Final):
    id: Id
    dc.validate(lambda self: self.id is not NotSpecified)

    _: dc.KW_ONLY

    #

    result: ta.Any = dc.field(default=NotSpecified)
    error: ta.Union['Error', type[NotSpecified]] = dc.field(default=NotSpecified)
    dc.validate(lambda self: is_not_specified(self.result) ^ is_not_specified(self.error))

    @property
    def is_result(self) -> bool:
        return not is_not_specified(self.result)

    @property
    def is_error(self) -> bool:
        return not is_not_specified(self.error)

    #

    jsonrpc: str = dc.field(default=VERSION)
    dc.validate(lambda self: self.jsonrpc == VERSION)


def result(id: Id, result: ta.Any) -> Response:  # noqa
    return Response(id, result=result)


##


@dc.dataclass(frozen=True)
@msh.update_fields_options(['data'], omit_if=is_not_specified)
class Error(lang.Final):
    code: int
    message: str
    data: ta.Any = NotSpecified


def error(id: Id, error: Error) -> Response:  # noqa
    return Response(id, error=error)


##


Message: ta.TypeAlias = Request | Response


def detect_message_type(dct: ta.Mapping[str, ta.Any]) -> type[Message]:
    if 'method' in dct:
        return Request
    else:
        return Response



from .api import (  # noqa
    SyncAsynclite as All,
)

from .base import (  # noqa
    SyncAsyncliteObject as Object,
    SyncAsyncliteApi as Api,
)

from .events import (  # noqa
    SyncAsyncliteEvent as Event,
    SyncAsyncliteEvents as Events,
)

from .identities import (  # noqa
    SyncAsyncliteIdentities as Identities,
)

from .locks import (  # noqa
    SyncAsyncliteLock as Lock,
    SyncAsyncliteLocks as Locks,
)

from .queues import (  # noqa
    SyncAsyncliteQueue as Queue,
    SyncAsyncliteQueues as Queues,
)

from .semaphores import (  # noqa
    SyncAsyncliteSemaphore as Semaphore,
    SyncAsyncliteSemaphores as Semaphores,
)

from .sleeps import (  # noqa
    SyncAsyncliteSleeps as Sleeps,
)



import dataclasses as dc
import typing as ta

from .maybes import Maybe


T = ta.TypeVar('T')


##


@dc.dataclass(init=False)
class Args:
    args: ta.Sequence[ta.Any]
    kwargs: ta.Mapping[str, ta.Any]

    def __init__(self, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__()

        self.args = args
        self.kwargs = kwargs

    def __bool__(self) -> bool:
        return bool(self.args) or bool(self.kwargs)

    def update(self, *args: ta.Any, **kwargs: ta.Any) -> 'Args':
        return Args(
            *self.args,
            *args,
            **{
                **self.kwargs,
                **kwargs,
            },
        )

    def map(self, fn: ta.Callable[[ta.Any], ta.Any]) -> 'Args':
        return Args(
            *[fn(a) for a in self.args],
            **{k: fn(v) for k, v in self.kwargs.items()},
        )

    def map_maybe(self, fn: ta.Callable[[ta.Any], Maybe[ta.Any]]) -> 'Args':
        return Args(
            *[n for a in self.args if (n := fn(a)).present],
            **{k: n for k, v in self.kwargs.items() if (n := fn(v)).present},
        )

    def __call__(self, fn: ta.Callable[..., T]) -> T:
        return fn(*self.args, **self.kwargs)

    @staticmethod
    def call(fn: ta.Callable[..., T], args: ta.Optional['Args']) -> T:
        if args is not None:
            return args(fn)
        else:
            return fn()

    EMPTY: ta.ClassVar['Args']


Args.EMPTY = Args()



import abc
import typing as ta

from .. import check
from .. import lang


K = ta.TypeVar('K')
V = ta.TypeVar('V')

K2 = ta.TypeVar('K2')
V2 = ta.TypeVar('V2')


##


class BiMap(ta.Mapping[K, V], lang.Abstract):
    @abc.abstractmethod
    def inverse(self) -> 'BiMap[V, K]':
        raise NotImplementedError


##


class _BaseBiMapImpl(BiMap[K, V], lang.Abstract):
    def __init__(
            self,
            dct: ta.Mapping[K, V],
            inverse: BiMap[V, K],
    ) -> None:
        super().__init__()

        self._dct = dct
        self._inverse = inverse

    def inverse(self) -> 'BiMap[V, K]':
        return self._inverse

    def __getitem__(self, key, /):
        return self._dct[key]

    def __len__(self) -> int:
        return len(self._dct)

    def __iter__(self):
        return iter(self._dct)

    def __repr__(self) -> str:
        return f'BiMap({self._dct!r})'


class _BiMapImpl(_BaseBiMapImpl[K, V]):
    def __init__(self, *args, **kwargs) -> None:
        dct: dict[K, V] = {}
        inv_dct: dict[V, K] = {}
        for k, v in lang.yield_dict_init(*args, **kwargs):
            check.not_in(k, dct)
            check.not_in(v, inv_dct)
            dct[k] = v
            inv_dct[v] = k

        super().__init__(dct, _BiMapImpl._Inverse(inv_dct, self))

    class _Inverse(_BaseBiMapImpl):
        pass


##


@ta.overload
def make_bi_map(**kwargs: V) -> BiMap[V, str]:
    ...


@ta.overload
def make_bi_map(dct: ta.Mapping[K, V]) -> BiMap[K, V]:
    ...


@ta.overload
def make_bi_map(items: ta.Iterable[tuple[K, V]]) -> BiMap[K, V]:
    ...


def make_bi_map(*args, **kwargs):
    return _BiMapImpl(*args, **kwargs)



# Based on pytest-trio, licensed under the MIT license, duplicated below.
#
#  https://github.com/python-trio/pytest-trio/tree/cd6cc14b061d34f35980e38c44052108ed5402d1
#
# The MIT License (MIT)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import collections.abc
import contextlib
import contextvars
import inspect
import typing as ta

from _pytest.outcomes import Skipped  # noqa
from _pytest.outcomes import XFailed  # noqa

from ..... import check
from ..... import lang
from .backends.base import AsyncsBackend
from .utils import is_coroutine_function


if ta.TYPE_CHECKING:
    import anyio.abc
else:
    anyio = lang.proxy_import('anyio', extras=['abc'])


##


CANARY: contextvars.ContextVar[ta.Any] = contextvars.ContextVar('pytest-omlish-asyncs canary')


class NURSERY_FIXTURE_PLACEHOLDER:  # noqa
    pass


##


class AsyncsTestContext:
    def __init__(self, backend: AsyncsBackend) -> None:
        super().__init__()

        self.backend = backend

        self.crashed = False

        # This holds cancel scopes for whatever setup steps are currently running -- initially it's the fixtures that
        # are in the middle of evaluating themselves, and then once fixtures are set up it's the test itself. Basically,
        # at any given moment, it's the stuff we need to cancel if we want to start tearing down our fixture DAG.
        self.active_cancel_scopes: set[anyio.CancelScope] = set()

        self.fixtures_with_errors: set[AsyncsFixture] = set()
        self.fixtures_with_cancel: set[AsyncsFixture] = set()

        self.error_list: list[BaseException] = []

    def crash(self, fixture: 'AsyncsFixture', exc: BaseException | None) -> None:
        if exc is None:
            self.fixtures_with_cancel.add(fixture)
        else:
            self.error_list.append(exc)
            self.fixtures_with_errors.add(fixture)

        self.crashed = True

        for cscope in self.active_cancel_scopes:
            cscope.cancel()


##


class AsyncsFixture:
    """
    Represent a fixture that need to be run in a async context to be resolved.

    The name is actually a misnomer, because we use it to represent the actual test itself as well, since the
    test is basically just a fixture with no dependents and no teardown.
    """

    def __init__(
            self,
            name: str,
            func: ta.Callable,
            pytest_kwargs: ta.Mapping[str, ta.Any],
            *,
            is_test: bool = False,
    ) -> None:
        super().__init__()

        self.name = name
        self._func = func
        self._pytest_kwargs = pytest_kwargs
        self._is_test = is_test
        self._teardown_done = anyio.Event()

        # These attrs are all accessed from other objects: Downstream users read this value.
        self.fixture_value: ta.Any = None

        # This event notifies downstream users that we're done setting up. Invariant: if this is set, then either
        # fixture_value is usable *or* test_ctx.crashed is True.
        self.setup_done = anyio.Event()

        # Downstream users *modify* this value, by adding their _teardown_done events to it, so we know who we need to
        # wait for before tearing down.
        self.user_done_events: set[anyio.Event] = set()

    def register_and_collect_dependencies(self) -> set['AsyncsFixture']:
        # Returns the set of all AsyncsFixtures that this fixture depends on, directly or indirectly, and sets up all
        # their user_done_events.
        deps = set()
        deps.add(self)
        for value in self._pytest_kwargs.values():
            if isinstance(value, AsyncsFixture):
                value.user_done_events.add(self._teardown_done)
                deps.update(value.register_and_collect_dependencies())
        return deps

    @contextlib.asynccontextmanager
    async def _fixture_manager(self, test_ctx: AsyncsTestContext) -> ta.AsyncIterator['anyio.abc.TaskGroup']:
        __tracebackhide__ = True

        try:
            async with anyio.create_task_group() as nursery_fixture:
                try:
                    yield nursery_fixture
                finally:
                    nursery_fixture.cancel_scope.cancel()

        except* (Skipped, XFailed):
            pass

        except* BaseException as exc:  # noqa
            test_ctx.crash(self, exc)

        finally:
            self.setup_done.set()
            self._teardown_done.set()

    async def run(
            self,
            test_ctx: AsyncsTestContext,
            contextvars_ctx: contextvars.Context,
    ) -> None:
        __tracebackhide__ = True

        await test_ctx.backend.install_context(contextvars_ctx)

        # Check that it worked, since technically trio doesn't *guarantee* that sleep(0) will actually yield.
        check.equal(CANARY.get(), 'in correct context')

        # This 'with' block handles the nursery fixture lifetime, the teardown_done event, and crashing the context if
        # there's an unhandled exception.
        async with self._fixture_manager(test_ctx) as nursery_fixture:
            # Resolve our kwargs
            resolved_kwargs: dict = {}
            for name, value in self._pytest_kwargs.items():
                if isinstance(value, AsyncsFixture):
                    await value.setup_done.wait()
                    if value.fixture_value is NURSERY_FIXTURE_PLACEHOLDER:
                        resolved_kwargs[name] = nursery_fixture
                    else:
                        resolved_kwargs[name] = value.fixture_value
                else:
                    resolved_kwargs[name] = value

            # If something's already crashed before we're ready to start, then there's no point in even setting up.
            if test_ctx.crashed:
                return

            # Run actual fixture setup step. If another fixture crashes while we're in the middle of setting up, we want
            # to be cancelled immediately, so we'll save an encompassing cancel scope where self._crash can find it.
            test_ctx.active_cancel_scopes.add(nursery_fixture.cancel_scope)
            if self._is_test:
                # Tests are exactly like fixtures, except that they to be
                # regular async functions.
                check.state(not self.user_done_events)
                func_value = None
                check.state(not test_ctx.crashed)
                await self._func(**resolved_kwargs)

            else:
                func_value = self._func(**resolved_kwargs)
                if isinstance(func_value, collections.abc.Coroutine):
                    self.fixture_value = await func_value
                elif inspect.isasyncgen(func_value):
                    self.fixture_value = await func_value.asend(None)
                elif isinstance(func_value, collections.abc.Generator):
                    self.fixture_value = func_value.send(None)
                else:
                    # Regular synchronous function
                    self.fixture_value = func_value

            # Now that we're done setting up, we don't want crashes to cancel us immediately; instead we want them to
            # cancel our downstream dependents, and then eventually let us clean up normally. So remove this from the
            # set of cancel scopes affected by self._crash.
            test_ctx.active_cancel_scopes.remove(nursery_fixture.cancel_scope)

            # self.fixture_value is ready, so notify users that they can continue. (Or, maybe we crashed and were
            # cancelled, in which case our users will check test_ctx.crashed and immediately exit, which is fine too.)
            self.setup_done.set()

            # Wait for users to be finished.
            #
            # At this point we're in a very strange state: if the fixture yielded inside a nursery or cancel scope, then
            # we are still "inside" that scope even though its with block is not on the stack. In particular this means
            # that if they get cancelled, then our waiting might get a Cancelled error, that we cannot really deal with
            # - it should get thrown back into the fixture generator, but pytest fixture generators don't work that way:
            #   https://github.com/python-trio/pytest-trio/issues/55
            # And besides, we can't start tearing down until all our users have finished.
            #
            # So if we get an exception here, we crash the context (which cancels the test and starts the cleanup
            # process), save any exception that *isn't* Cancelled (because if its Cancelled then we can't route it to
            # the right place, and anyway the teardown code will get it again if it matters), and then use a shield to
            # keep waiting for the teardown to finish without having to worry about cancellation.
            yield_outcome: lang.Outcome = lang.Value(None)
            try:
                for event in self.user_done_events:
                    await event.wait()

            except BaseException as exc:  # noqa
                check.isinstance(exc, anyio.get_cancelled_exc_class())
                yield_outcome = lang.Error(exc)
                test_ctx.crash(self, None)
                with anyio.CancelScope(shield=True):
                    for event in self.user_done_events:
                        await event.wait()

            # Do our teardown
            if inspect.isasyncgen(func_value):
                try:
                    await yield_outcome.asend(func_value)
                except StopAsyncIteration:
                    pass
                else:
                    raise RuntimeError('too many yields in fixture')

            elif isinstance(func_value, collections.abc.Generator):
                try:
                    yield_outcome.send(func_value)
                except StopIteration:
                    pass
                else:
                    raise RuntimeError('too many yields in fixture')


##


def is_asyncs_fixture(
        func: ta.Callable,
        coerce_async: bool,
        kwargs: ta.Mapping[str, ta.Any],
) -> bool:
    if coerce_async and (is_coroutine_function(func) or inspect.isasyncgenfunction(func)):
        return True

    if any(isinstance(value, AsyncsFixture) for value in kwargs.values()):
        return True

    return False



from ... import dataclasses as dc
from ... import lang
from .elements import Element
from .elements import Elements


##


@dc.dataclass(frozen=True)
class TableDef(lang.Final):
    name: str
    elements: Elements


def table_def(
        name: str,
        *elements: Element,
) -> TableDef:
    return TableDef(
        name,
        Elements(elements),
    )



"""
TODO:
 - dataclasses
"""
import argparse
import os
import re
import resource
import struct
import sys
import typing as ta

from .. import iterators as it
from .. import lang
from ..formats import json
from ..logs import all as logs
from ..os.sizes import PAGE_SIZE
from .procstats import ProcStats


log = logs.get_module_logger(globals())


PidLike = int | str


##


RLIMIT_RESOURCES = {
    getattr(resource, k): k
    for k in dir(resource)
    if k.startswith('RLIMIT_')
}


def parse_size(s: str) -> int:
    if ' ' not in s:
        return int(s)
    us = {'kB': 1024, 'mB': 1024 * 1024}
    v, u = s.split()
    return int(v) * us[u]


class ProcStat(lang.Namespace, lang.Final):
    PID = 0
    COMM = 1
    STATE = 2
    PPID = 3
    PGRP = 4
    SESSION = 5
    TTY_NR = 6
    TPGID = 7
    FLAGS = 8
    MINFLT = 9
    CMINFLT = 10
    MAJFLT = 11
    CMAJFLT = 12
    UTIME = 13
    STIME = 14
    CUTIME = 15
    CSTIME = 16
    PRIORITY = 17
    NICE = 18
    NUM_THREADS = 19
    ITREALVALUE = 20
    STARTTIME = 21
    VSIZE = 22
    RSS = 23
    RSSLIM = 24
    STARTCODE = 25
    ENDCODE = 26
    STARTSTACK = 27
    KSTKESP = 28
    KSTKEIP = 29
    SIGNAL = 30
    BLOCKED = 31
    SIGIGNORE = 32
    SIGCATCH = 33
    WCHAN = 34
    NSWAP = 35
    CNSWAP = 36
    EXIT_SIGNAL = 37
    PROCESSOR = 38
    RT_PRIORITY = 39
    POLICY = 40
    DELAYACCT_BLKIO_TICKS = 41
    GUEST_TIME = 42
    CGUEST_TIME = 43
    START_DATA = 44
    END_DATA = 45
    START_BRK = 46
    ARG_START = 47
    ARG_END = 48
    ENV_START = 49
    ENV_END = 50
    EXIT_CODE = 51


def _check_linux() -> None:
    if sys.platform != 'linux':
        raise OSError


##


def get_process_stats(pid: PidLike = 'self') -> list[str]:
    """http://man7.org/linux/man-pages/man5/proc.5.html -> /proc/[pid]/stat"""

    _check_linux()
    with open(f'/proc/{pid}/stat') as f:
        buf = f.read()
    l, _, r = buf.rpartition(')')
    pid, _, comm = l.partition('(')
    return [pid.strip(), comm, *r.strip().split(' ')]


def get_process_procstats(pid: int | None = None) -> ProcStats:
    st = get_process_stats('self' if pid is None else pid)
    return ProcStats(
        pid=int(st[ProcStat.PID]),

        rss=int(st[ProcStat.RSS]),
    )


##


def get_process_chain(pid: PidLike = 'self') -> list[tuple[int, str]]:
    _check_linux()
    lst = []
    while pid:
        process_stats = get_process_stats(pid)
        lst.append((int(process_stats[ProcStat.PID]), process_stats[ProcStat.COMM]))
        pid = int(process_stats[ProcStat.PPID])
    return lst


def get_process_start_time(pid: PidLike = 'self') -> int:
    """https://stackoverflow.com/questions/2598145/how-to-retrieve-the-process-start-time-or-uptime-in-python"""

    _check_linux()
    hz = os.sysconf(os.sysconf_names['SC_CLK_TCK'])
    with open('/proc/stat') as f:
        system_stats = f.readlines()
    for line in system_stats:
        if line.startswith('btime'):
            boot_timestamp = int(line.split()[1])
            break
    else:
        raise ValueError
    process_stats = get_process_stats(pid)
    age_from_boot_jiffies = int(process_stats[ProcStat.STARTTIME])
    age_from_boot_timestamp = age_from_boot_jiffies // hz
    return boot_timestamp + age_from_boot_timestamp


def get_process_rss(pid: PidLike = 'self') -> int:
    return int(get_process_stats(pid)[ProcStat.RSS])


def set_process_oom_score_adj(score: str, pid: PidLike = 'self') -> None:
    _check_linux()
    with open(f'/proc/{pid}/oom_score_adj', 'w') as f:
        f.write(str(score))


##


MAP_LINE_RX = re.compile(
    r'^'
    r'(?P<address>[A-Fa-f0-9]+)-(?P<end_address>[A-Fa-f0-9]+)\s+'
    r'(?P<permissions>\S+)\s+'
    r'(?P<offset>[A-Fa-f0-9]+)\s+'
    r'(?P<device>\S+)\s+'
    r'(?P<inode>\d+)\s+'
    r'(?P<path>.*)'
    r'$',
)


def get_process_maps(pid: PidLike = 'self', sharing: bool = False) -> ta.Iterator[dict[str, ta.Any]]:
    """http://man7.org/linux/man-pages/man5/proc.5.html -> /proc/[pid]/maps"""

    _check_linux()
    with open(f'/proc/{pid}/{"smaps" if sharing else "maps"}') as map_file:
        while True:
            line = map_file.readline()
            if not line:
                break
            m = MAP_LINE_RX.match(line)
            if not m:
                raise ValueError(line)
            address = int(m.group('address'), 16)
            end_address = int(m.group('end_address'), 16)
            d = {
                'address': address,
                'end_address': end_address,
                'size': end_address - address,
                'permissions': [x for x in m.group('permissions') if x != '-'],
                'offset': int(m.group('offset'), 16),
                'device': m.group('device'),
                'inode': int(m.group('inode')),
                'path': m.group('path'),
            }
            if sharing:
                s: dict[str, ta.Any] = {}
                while True:
                    line = map_file.readline()
                    k, v = line.split(':')
                    if k.lower() == 'vmflags':
                        break
                    s[k.lower()] = parse_size(v.strip())
                _, v = line.split(':')
                s['vmflags'] = [p for p in [j.strip() for j in v.split(' ')] if p]
                d['sharing'] = s
            yield d


##


PAGEMAP_KEYS = (
    'address',
    'pfn',
    'swap_type',
    'swap_offset',
    'pte_soft_dirty',
    'file_page_or_shared_anon',
    'page_swapped',
    'page_present',
)


def get_process_range_pagemaps(start: int, end: int, pid: PidLike = 'self') -> ta.Iterator[dict[str, int]]:
    """https://www.kernel.org/doc/Documentation/vm/pagemap.txt"""

    _check_linux()
    offset = (start // PAGE_SIZE) * 8
    npages = ((end - start) // PAGE_SIZE)
    size = npages * 8
    with open(f'/proc/{pid}/pagemap', 'rb') as pagemap_file:
        pagemap_file.seek(offset)
        pagemap_buf = pagemap_file.read(size)
    if not pagemap_buf:
        return
    _struct_unpack = struct.unpack
    for pagenum in range(npages):
        [packed] = _struct_unpack('Q', pagemap_buf[pagenum * 8:(pagenum + 1) * 8])
        yield {
            'address': start + (pagenum * PAGE_SIZE),
            'pfn': (packed & ((1 << (54 + 1)) - 1)),
            'swap_type': (packed & ((1 << (4 + 1)) - 1)),
            'swap_offset': (packed & ((1 << (54 + 1)) - 1)) >> 5,
            'pte_soft_dirty': ((packed >> 55) & 1) > 0,
            'file_page_or_shared_anon': ((packed >> 61) & 1) > 0,
            'page_swapped': ((packed >> 62) & 1) > 0,
            'page_present': ((packed >> 63) & 1) > 0,
        }


def get_process_pagemaps(pid: PidLike = 'self') -> ta.Iterator[dict[str, int]]:
    _check_linux()
    for m in get_process_maps(pid):
        yield from get_process_range_pagemaps(m['address'], m['end_address'], pid)


##


def _dump_cmd(args: ta.Any) -> None:
    total = 0
    dirty_total = 0
    for m in get_process_maps(args.pid, sharing=True):
        total += m['sharing']['rss']
        sys.stdout.write(json.dumps({'map': m}))
        sys.stdout.write('\n')
        for pm in get_process_range_pagemaps(m['address'], m['end_address'], args.pid):
            if pm['pte_soft_dirty']:
                dirty_total += PAGE_SIZE
            sys.stdout.write(json.dumps({'page': tuple(pm[k] for k in PAGEMAP_KEYS)}))
            sys.stdout.write('\n')
    dct = {
        'total': total,
        'dirty_total': dirty_total,
    }
    sys.stdout.write(json.dumps(dct))
    sys.stdout.write('\n')


def _cmp_cmd(args: ta.Any) -> None:
    if len(args.pids) == 1:
        [rpid] = args.pids
        lpid = get_process_chain(rpid)[1][0]
    elif len(args.pids) == 2:
        lpid, rpid = args.pids
    else:
        raise TypeError('Invalid arguments')

    def g(pid: int) -> ta.Iterator[dict[str, int]]:
        for m in get_process_maps(pid, sharing=True):
            yield from get_process_range_pagemaps(m['address'], m['end_address'], pid)

    lpms, rpms = (g(pid) for pid in (lpid, rpid))

    l_pages = 0
    r_pages = 0
    c_pages = 0
    for _, ps in it.merge_on(lambda pm: pm['address'], lpms, rpms):
        l, r = it.expand_indexed_pairs(ps, None, width=2)
        if l is not None and r is None:
            l_pages += 1
        elif l is None and r is not None:
            r_pages += 1
        elif l['pfn'] != r['pfn']:  # type: ignore
            c_pages += 1
        else:
            continue
        if not args.quiet:
            sys.stdout.write(json.dumps([l, r]))
            sys.stdout.write('\n')
    l_pages += c_pages
    r_pages += c_pages
    dct = {
        'l_pages': l_pages,
        'l_bytes': l_pages * PAGE_SIZE,
        'r_pages': r_pages,
        'r_bytes': r_pages * PAGE_SIZE,
        'c_pages': c_pages,
        'c_bytes': c_pages * PAGE_SIZE,
    }
    sys.stdout.write(json.dumps(dct))
    sys.stdout.write('\n')


def _main() -> None:
    _check_linux()

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument('-q', '--quiet', action='store_true')
    arg_subparsers = arg_parser.add_subparsers()

    dump_arg_parser = arg_subparsers.add_parser('dump')
    dump_arg_parser.add_argument('pid', type=int)
    dump_arg_parser.set_defaults(func=_dump_cmd)

    cmp_arg_parser = arg_subparsers.add_parser('cmp')
    cmp_arg_parser.add_argument('pids', type=int, nargs='*')
    cmp_arg_parser.set_defaults(func=_cmp_cmd)

    args = arg_parser.parse_args()
    if not hasattr(args, 'func'):
        arg_parser.print_help()
    else:
        args.func(args)


# @omlish-manifest
_CLI_MODULE = {'!omdev.cli.types.CliModule': {
    'name': 'procfs',
    'module': __name__,
}}


if __name__ == '__main__':
    _main()



"""
A 'guard function' (`GuardFn`) is a callable that simultaneously tests whether it can handle a given set of arguments
and, if so, returns a zero-argument thunk encapsulating the actual computation to perform. If the guard cannot handle
the arguments it returns `None`. The defining characteristic -- and the key difference from a simple predicate -- is
that this single call combines both the applicability test and the argument binding: any work done during the test (type
narrowing, parsing, pattern matching, preliminary computation) is naturally captured in the returned thunk's closure,
available when and if that thunk is eventually called.

This design is related to but not directly equivalent to Scala's `PartialFunction`
(https://www.scala-lang.org/api/current/scala/PartialFunction.html). Both represent functions defined only over a subset
of their domain. A Scala `PartialFunction` separates applicability testing (`isDefinedAt`) from application (`apply`),
forcing any shared computation to be performed twice. A `GuardFn` merges these into a single call: the returned thunk
already has the arguments bound from the guard-test invocation, so no work need be repeated. Whether the returned thunk
is reusable -- i.e. whether it may be called zero, one, or more times -- is intentionally outside the scope of this
abstraction and is left to the use case.

Also provided: `MultiGuardFn` for composing multiple guards with optional strict single-match enforcement;
`GuardFnMethod` and `ImmediateGuardFnMethod` descriptors for class-method-style dispatch with inheritance-aware handler
registration, exposed via the `method()` and `immediate_method()` decorators.
"""
import abc
import functools
import operator
import typing as ta

from .. import check
from .. import collections as col
from .. import lang


T = ta.TypeVar('T')
T_co = ta.TypeVar('T_co', covariant=True)
U = ta.TypeVar('U')
P = ta.ParamSpec('P')


##


class GuardFn(ta.Protocol[P, T_co]):
    def __get__(self, instance, owner=None): ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T_co] | None: ...


##


@ta.final
class DumbGuardFn(ta.Generic[P, T]):
    def __init__(self, fn: ta.Callable[P, T]) -> None:
        self._fn = fn

    def __get__(self, instance, owner=None):
        return DumbGuardFn(self._fn.__get__(instance, owner))  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T]:
        return functools.partial(self._fn, *args, **kwargs)


dumb = DumbGuardFn


##


class AmbiguousGuardFnError(Exception):
    pass


@ta.final
class MultiGuardFn(ta.Generic[P, T]):
    def __init__(
            self,
            *children: GuardFn[P, T],
            default: GuardFn[P, T] | None = None,
            strict: bool = False,
    ) -> None:
        self._children, self._default, self._strict = children, default, strict

    lang.attr_ops(lambda self: (
        self._children,
        self._default,
        self._strict,
    )).install(locals())

    def __get__(self, instance, owner=None):
        return MultiGuardFn(*map(operator.methodcaller('__get__', instance, owner), self._children), default=self._default.__get__(instance, owner) if self._default is not None else None, strict=self._strict)  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T] | None:
        matches = []
        for c in self._children:
            if (m := c(*args, **kwargs)) is not None:
                if not self._strict:
                    return m
                matches.append(m)
        if not matches:
            if (dfl := self._default) is not None:
                return dfl(*args, **kwargs)
            else:
                return None
        elif len(matches) > 1:
            raise AmbiguousGuardFnError
        else:
            return matches[0]


multi = MultiGuardFn


##


class _BaseGuardFnMethod(lang.Abstract, ta.Generic[P, T]):
    def __init__(
            self,
            *,
            strict: bool = False,
            requires_override: bool = False,
            instance_cache: bool = False,
            default: GuardFn[P, T] | None = None,
    ) -> None:
        super().__init__()

        self._strict = strict
        self._instance_cache = instance_cache
        self._default = default

        self._registry: col.AttrRegistry[ta.Callable, None] = col.AttrRegistry(
            requires_override=requires_override,
        )

        self._cache: col.AttrRegistryCache[ta.Callable, None, MultiGuardFn] = col.WeakAttrRegistryCache(
            self._registry,
            self._prepare,
        )

    _owner: type | None = None
    _name: str | None = None

    def __set_name__(self, owner, name):
        if self._owner is None:
            self._owner = owner
        if self._name is None:
            self._name = name

    def register(self, fn: U) -> U:
        check.callable(fn)
        self._registry.register(ta.cast(ta.Callable, fn), None)
        return fn

    def _prepare(self, instance_cls: type, collected: ta.Mapping[str, tuple[ta.Callable, None]]) -> MultiGuardFn:
        return MultiGuardFn(
            *[getattr(instance_cls, a) for a in collected],
            default=self._default,
            strict=self._strict,
        )

    @abc.abstractmethod
    def _bind(self, instance, owner):
        raise NotImplementedError

    def __get__(self, instance, owner=None):
        if instance is None:
            return self

        if self._instance_cache:
            try:
                return instance.__dict__[self._name]
            except KeyError:
                pass

        bound = self._bind(instance, owner)

        if self._instance_cache:
            instance.__dict__[self._name] = bound

        return bound

    def _call(self, *args, **kwargs):
        instance, *rest = args
        return self.__get__(instance)(*rest, **kwargs)

#


@ta.final
class GuardFnMethod(_BaseGuardFnMethod[P, T]):
    def _bind(self, instance, owner):
        return self._cache.get(type(instance)).__get__(instance, owner)  # noqa

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> ta.Callable[[], T] | None:
        return self._call(*args, **kwargs)


def method(
        *,
        strict: bool = False,
        requires_override: bool = False,
        instance_cache: bool = False,
        default: bool = False,
) -> ta.Callable[[ta.Callable[P, T]], GuardFnMethod[P, T]]:  # noqa
    def inner(fn):
        return GuardFnMethod(
            strict=strict,
            requires_override=requires_override,
            instance_cache=instance_cache,
            default=fn if default else None,
        )

    return inner


#


@ta.final
class ImmediateGuardFnMethod(_BaseGuardFnMethod[P, T]):
    def _bind(self, instance, owner):
        gf = self._cache.get(type(instance)).__get__(instance, owner)  # noqa

        def inner(*args, **kwargs):
            return gf(*args, **kwargs)()  # Note: cannot be None due to non-optional default

        return inner

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self._call(*args, **kwargs)


def immediate_method(
        *,
        strict: bool = False,
        requires_override: bool = False,
        instance_cache: bool = False,
) -> ta.Callable[[ta.Callable[P, T]], ImmediateGuardFnMethod[P, T]]:  # noqa
    def inner(fn):
        return ImmediateGuardFnMethod(
            strict=strict,
            requires_override=requires_override,
            instance_cache=instance_cache,
            default=(lambda *args, **kwargs: lambda: fn(*args, **kwargs)),
        )

    return inner



# ruff: noqa: N802 N803 UP006 UP007 UP045
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# --------------------------------------------
#
# 1. This LICENSE AGREEMENT is between the Python Software Foundation ("PSF"), and the Individual or Organization
# ("Licensee") accessing and otherwise using this software ("Python") in source or binary form and its associated
# documentation.
#
# 2. Subject to the terms and conditions of this License Agreement, PSF hereby grants Licensee a nonexclusive,
# royalty-free, world-wide license to reproduce, analyze, test, perform and/or display publicly, prepare derivative
# works, distribute, and otherwise use Python alone or in any derivative version, provided, however, that PSF's License
# Agreement and PSF's notice of copyright, i.e., "Copyright (c) 2001-2024 Python Software Foundation; All Rights
# Reserved" are retained in Python alone or in any derivative version prepared by Licensee.
#
# 3. In the event Licensee prepares a derivative work that is based on or incorporates Python or any part thereof, and
# wants to make the derivative work available to others as provided herein, then Licensee hereby agrees to include in
# any such work a brief summary of the changes made to Python.
#
# 4. PSF is making Python available to Licensee on an "AS IS" basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES,
# EXPRESS OR IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND DISCLAIMS ANY REPRESENTATION OR WARRANTY
# OF MERCHANTABILITY OR FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT INFRINGE ANY THIRD PARTY
# RIGHTS.
#
# 5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL
# DAMAGES OR LOSS AS A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON, OR ANY DERIVATIVE THEREOF, EVEN IF
# ADVISED OF THE POSSIBILITY THEREOF.
#
# 6. This License Agreement will automatically terminate upon a material breach of its terms and conditions.
#
# 7. Nothing in this License Agreement shall be deemed to create any relationship of agency, partnership, or joint
# venture between PSF and Licensee.  This License Agreement does not grant permission to use PSF trademarks or trade
# name in a trademark sense to endorse or promote products or services of Licensee, or any third party.
#
# 8. By copying, installing or otherwise using Python, Licensee agrees to be bound by the terms and conditions of this
# License Agreement.
import abc
import contextvars
import functools
import inspect
import typing as ta
import unittest
import warnings

from ...lite.asyncs import sync_await
from ...lite.check import check


##


class IsolatedAsyncTestCaseAsyncRunner(abc.ABC):
    def setup(self) -> None:
        pass

    @abc.abstractmethod
    def run(self, fn: ta.Callable) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        raise NotImplementedError


class IsolatedAsyncTestCase(unittest.TestCase):
    def __init__(self, methodName='runTest'):
        super().__init__(methodName)

        self._asyncRunner: ta.Optional[IsolatedAsyncTestCaseAsyncRunner] = None
        self._asyncTestContext = contextvars.copy_context()

    async def asyncSetUp(self):
        pass

    async def asyncTearDown(self):
        pass

    def addAsyncCleanup(self, func, /, *args, **kwargs):
        # A trivial trampoline to addCleanup() the function exists because it has a different semantics and signature:
        # addCleanup() accepts regular functions but addAsyncCleanup() accepts coroutines
        #
        # We intentionally don't add inspect.iscoroutinefunction() check for func argument because there is no way to
        # check for async function reliably:
        # 1. It can be "async def func()" itself
        # 2. Class can implement "async def __call__()" method
        # 3. Regular "def func()" that returns awaitable object
        self.addCleanup(*(func, *args), **kwargs)

    async def enterAsyncContext(self, cm):
        """
        Enters the supplied asynchronous context manager.

        If successful, also adds its __aexit__ method as a cleanup function and returns the result of the __aenter__
        method.
        """

        # We look up the special methods on the type to match the with statement.
        cls = type(cm)
        try:
            enter = cls.__aenter__
            exit = cls.__aexit__  # noqa
        except AttributeError:
            raise TypeError(
                f"'{cls.__module__}.{cls.__qualname__}' object does not support the asynchronous context manager "
                f"protocol",
            ) from None
        result = await enter(cm)
        self.addAsyncCleanup(exit, cm, None, None, None)
        return result

    def _callSetUp(self):
        # Force loop to be initialized and set as the current loop so that setUp functions can use get_event_loop() and
        # get the correct loop instance.
        check.not_none(self._asyncRunner).setup()
        self._asyncTestContext.run(self.setUp)
        self._callAsync(self.asyncSetUp)

    def _callTestMethod(self, method):
        if self._callMaybeAsync(method) is not None:
            warnings.warn(
                f'It is deprecated to return a value that is not None from a test case ({method})',
                DeprecationWarning,
                stacklevel=4,
            )

    def _callTearDown(self):
        self._callAsync(self.asyncTearDown)
        self._asyncTestContext.run(self.tearDown)

    def _callCleanup(self, function, *args, **kwargs):
        self._callMaybeAsync(function, *args, **kwargs)

    def _callAsync(self, func, /, *args, **kwargs):
        runner = check.not_none(self._asyncRunner)
        check.state(inspect.iscoroutinefunction(func))
        if args or kwargs:
            func = functools.partial(func, *args, **kwargs)
        func = functools.partial(self._asyncTestContext.run, func)
        return runner.run(func)

    def _callMaybeAsync(self, func, /, *args, **kwargs):
        if inspect.iscoroutinefunction(func):
            return self._callAsync(func, *args, **kwargs)
        else:
            return self._asyncTestContext.run(func, *args, **kwargs)

    def _createAsyncRunner(self):
        raise NotImplementedError

    def _setupAsyncRunner(self):
        check.none(self._asyncRunner)
        runner = self._createAsyncRunner()
        self._asyncRunner = runner

    def _tearDownAsyncRunner(self):
        if (runner := self._asyncRunner) is not None:
            runner.close()
            self._asyncRunner = None

    def run(self, result=None):
        self._setupAsyncRunner()
        try:
            return super().run(result)
        finally:
            self._tearDownAsyncRunner()

    def debug(self):
        self._setupAsyncRunner()
        super().debug()
        self._tearDownAsyncRunner()

    def __del__(self):
        if self._asyncRunner is not None:
            self._tearDownAsyncRunner()


##


class AsyncioIsolatedAsyncTestCaseAsyncRunner(IsolatedAsyncTestCaseAsyncRunner):
    # 3.8 has no asyncio.Runner, cannot use :|
    #
    # def __init__(self, *, loop_factory=None):
    #     import asyncio
    #     self._runner = asyncio.Runner(debug=True, loop_factory=loop_factory)
    #
    # def setup(self) -> None:
    #     self._runner.get_loop()
    #
    # def run(self, obj: ta.Any, *, context: ta.Optional[contextvars.Context] = None) -> ta.Any:
    #     return self._runner.run(obj, context=context)
    #
    # def close(self) -> None:
    #     self._runner.close()

    def __init__(self, *, loop_factory=None):
        pass

    _loop: ta.Any = None

    _task: ta.Any
    _queue: ta.Any

    def setup(self) -> None:
        import asyncio

        check.none(self._loop)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.set_debug(True)
        self._loop = loop

        fut = loop.create_future()
        self._task = loop.create_task(self._runner_main(fut))
        loop.run_until_complete(fut)

    def run(self, fn: ta.Callable) -> ta.Any:
        fut = self._loop.create_future()
        self._queue.put_nowait((fut, fn))
        return self._loop.run_until_complete(fut)

    def close(self) -> None:
        import asyncio

        check.not_none(self._loop)
        loop = self._loop
        self._loop = None

        self._queue.put_nowait(None)
        loop.run_until_complete(self._queue.join())

        try:
            # cancel all tasks
            to_cancel = asyncio.all_tasks(loop)
            if not to_cancel:
                return

            for task in to_cancel:
                task.cancel()

            loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

            for task in to_cancel:
                if task.cancelled():
                    continue

                if task.exception() is not None:
                    loop.call_exception_handler({
                        'message': 'unhandled exception during test shutdown',
                        'exception': task.exception(),
                        'task': task,
                    })

            # shutdown asyncgens
            loop.run_until_complete(loop.shutdown_asyncgens())

        finally:
            asyncio.set_event_loop(None)
            loop.close()

    async def _runner_main(self, fut):
        import asyncio

        queue: asyncio.Queue = asyncio.Queue()
        self._queue = queue

        fut.set_result(None)

        while True:
            query = await queue.get()
            queue.task_done()

            if query is None:
                return

            fut, fn = query
            try:
                ret = await fn()

                if not fut.cancelled():
                    fut.set_result(ret)

            except (SystemExit, KeyboardInterrupt):
                raise

            except (BaseException, asyncio.CancelledError) as ex:  # noqa
                if not fut.cancelled():
                    fut.set_exception(ex)


class AsyncioIsolatedAsyncTestCase(IsolatedAsyncTestCase):
    def _createAsyncRunner(self):
        return AsyncioIsolatedAsyncTestCaseAsyncRunner()


##


class SyncIsolatedAsyncTestCaseAsyncRunner(IsolatedAsyncTestCaseAsyncRunner):
    def __init__(self):
        pass

    def run(self, fn: ta.Callable) -> ta.Any:
        return sync_await(fn())

    def close(self) -> None:
        pass


class SyncIsolatedAsyncTestCase(IsolatedAsyncTestCase):
    def _createAsyncRunner(self):
        return SyncIsolatedAsyncTestCaseAsyncRunner()


##


class AnyioIsolatedAsyncTestCaseAsyncRunner(IsolatedAsyncTestCaseAsyncRunner):
    def __init__(self, **kwargs):
        import anyio.from_thread
        self._cm = anyio.from_thread.start_blocking_portal(**kwargs)
        self._portal = self._cm.__enter__()

    def run(self, fn: ta.Callable) -> ta.Any:
        async def inner():
            return await fn()

        return self._portal.call(inner)

    def close(self) -> None:
        self._cm.__exit__(None, None, None)


class AnyioIsolatedAsyncTestCase(IsolatedAsyncTestCase):
    def _createAsyncRunner(self):
        return AnyioIsolatedAsyncTestCaseAsyncRunner()



import abc
import dataclasses as dc
import types
import typing as ta

from .. import c3
from .. import lang
from .ops import get_concrete_type
from .ops import to_annotation
from .types import Any
from .types import Generic
from .types import Literal
from .types import NewType
from .types import Type
from .types import Union
from .types import get_params
from .types import type_


if ta.TYPE_CHECKING:
    from ..collections import cache
else:
    cache = lang.proxy_import('.collections.cache', __package__)


GenericBasesMap: ta.TypeAlias = ta.Mapping[Type, tuple[Type, ...]]


##


def get_type_var_replacements(ty: Type) -> ta.Mapping[ta.TypeVar, Type]:
    if isinstance(ty, Generic):
        return dict(zip(ty.params, ty.args))

    return {}


def replace_type_vars(
        ty: Type,
        rpl: ta.Mapping[ta.TypeVar, Type],
        *,
        update_aliases: bool = False,
) -> Type:
    def rec(cur):
        if isinstance(cur, (type, NewType, Any, Literal)):
            return cur

        if isinstance(cur, Generic):
            args = tuple(rec(a) for a in cur.args)
            if update_aliases:
                obj = cur.obj
                if (og := ta.get_origin(obj)) is ta.Generic:
                    # FIXME: None? filter_typing_generic in get_generic_bases?
                    pass
                elif (ops := get_params(og)):
                    if len(ops) != len(args):
                        raise RuntimeError
                    nargs = [to_annotation(a) for a in args]
                    obj = og[*nargs]
            else:
                obj = None
            return dc.replace(cur, args=args, obj=obj)

        if isinstance(cur, Union):
            return Union(frozenset(rec(e) for e in cur.args))

        if isinstance(cur, ta.TypeVar):
            return rpl[cur]

        raise TypeError(cur)

    return rec(ty)


##


_DEFAULT_SIMPLE_GENERIC_BASES: dict[Type, tuple[Type, ...]] = {}
DEFAULT_SIMPLE_GENERIC_BASES: GenericBasesMap = _DEFAULT_SIMPLE_GENERIC_BASES


class GenericSubstitution:
    def __init__(
            self,
            *,
            update_aliases: bool = False,
            cache_size: int = 0,  # FIXME: ta.Generic isn't weakrefable..
            simple_generic_bases: GenericBasesMap | None = None,
    ) -> None:
        super().__init__()

        self._update_aliases = update_aliases
        self._simple_generic_bases = simple_generic_bases

        if cache_size > 0:
            self.get_generic_bases = cache.cache(weak_keys=True, max_size=cache_size)(self.get_generic_bases)  # type: ignore  # noqa
            self.generic_mro = cache.cache(weak_keys=True, max_size=cache_size)(self.generic_mro)  # type: ignore

    def get_generic_bases(self, ty: Type) -> tuple[Type, ...]:
        if (sgm := self._simple_generic_bases) is not None:
            if (sgb := sgm.get(ty)) is not None:
                return sgb

        if (cty := get_concrete_type(ty)) is not None:
            rpl = get_type_var_replacements(ty)
            ret: list[Type] = []
            obs = types.get_original_bases(cty)
            for b in obs:
                bty = type_(b)
                if isinstance(bty, Generic) and isinstance(b, type):
                    # FIXME: throws away relative types, but can't use original vars as they're class-contextual
                    bty = type_(b[*((ta.Any,) * len(bty.params))])  # type: ignore
                rty = replace_type_vars(bty, rpl, update_aliases=self._update_aliases)
                ret.append(rty)
            return tuple(ret)

        return ()

    def generic_mro(self, obj: ta.Any) -> list[Type]:
        mro = c3.mro(
            type_(obj),
            get_bases=lambda t: self.get_generic_bases(t),
            is_subclass=lambda l, r: issubclass(get_concrete_type(l), get_concrete_type(r)),  # type: ignore
        )
        return [ty for ty in mro if get_concrete_type(ty) is not ta.Generic]


DEFAULT_GENERIC_SUBSTITUTION = GenericSubstitution(
    simple_generic_bases=DEFAULT_SIMPLE_GENERIC_BASES,
)

get_generic_bases = DEFAULT_GENERIC_SUBSTITUTION.get_generic_bases
generic_mro = DEFAULT_GENERIC_SUBSTITUTION.generic_mro

ALIAS_UPDATING_GENERIC_SUBSTITUTION = GenericSubstitution(
    update_aliases=True,
    simple_generic_bases=DEFAULT_SIMPLE_GENERIC_BASES,
)


##


_DEFAULT_SIMPLE_GENERIC_BASE_LIST: ta.Sequence[Type] = [
    object,
    abc.ABC,
    lang.Abstract,
    lang.Sealed,
    lang.PackageSealed,
    lang.Final,
]

_DEFAULT_SIMPLE_GENERIC_BASES.update({
    b: get_generic_bases(b)
    for b in _DEFAULT_SIMPLE_GENERIC_BASE_LIST
})



class NamespaceClass:
    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    def __init_subclass__(cls, **kwargs):  # noqa
        super().__init_subclass__(**kwargs)

        if any(issubclass(b, NamespaceClass) and b is not NamespaceClass for b in cls.__bases__):
            raise TypeError



import typing as ta

from ... import reflect as rfl
from ...funcs import guard as gfs
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory


##


class MultiMarshalerFactory(MarshalerFactory):
    def __init__(
            self,
            *facs: MarshalerFactory,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._facs = facs
        self._mgf = gfs.multi(*[f.make_marshaler for f in self._facs], strict=strict)

    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        return self._mgf(ctx, rty)


class MultiUnmarshalerFactory(UnmarshalerFactory):
    def __init__(
            self,
            *facs: UnmarshalerFactory,
            strict: bool = False,
    ) -> None:
        super().__init__()

        self._facs = facs
        self._mgf = gfs.multi(*[f.make_unmarshaler for f in self._facs], strict=strict)

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        return self._mgf(ctx, rty)



from ... import dataclasses as _dc


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .adapters import (  # noqa
    Adapter,

    HasAdapter,
)

from .asquery import (  # noqa
    AsQueryContext,
    as_query,
    as_query_,
)

from .columns import (  # noqa
    Column,
    Columns,
)

from .core import (  # noqa
    AnyRows,
    Rows,
    AsyncRows,

    AnyTransaction,
    Transaction,
    AsyncTransaction,

    AnyConn,
    Conn,
    AsyncConn,

    AnyDb,
    Db,
    AsyncDb,
)

from .dbapi import (  # noqa
    DbapiRows,
    DbapiConn,
    DbapiDb,
    DbapiAdapter,
)

from .errors import (  # noqa
    Error,

    ColumnError,
    DuplicateColumnNameError,
    MismatchedColumnCountError,

    QueryError,
)

from .queriers import (  # noqa
    AnyQuerier,
    Querier,
    AsyncQuerier,
)

from .querierfuncs import (  # noqa
    sync_exec,
    async_exec,
    exec,  # noqa

    sync_query,
    async_query,
    query,

    sync_query_all,
    async_query_all,
    query_all,

    sync_query_first,
    async_query_first,
    query_first,

    sync_query_opt_first,
    async_query_opt_first,
    query_opt_first,

    sync_query_one,
    async_query_one,
    query_one,

    sync_query_opt_one,
    async_query_opt_one,
    query_opt_one,

    sync_query_scalar,
    async_query_scalar,
    query_scalar,

    sync_query_maybe_scalar,
    async_query_maybe_scalar,
    query_maybe_scalar,
)

from .queries import (  # noqa
    QueryMode,

    Queryable,
    Query,
)

from .rows import (  # noqa
    Row,
)


##


from ... import lang as _lang  # noqa

_lang.register_conditional_import('..queries', '._queries', __package__)



# ruff: noqa: PYI034 UP006 UP007 UP045
# @omlish-lite
import functools
import logging
import time
import typing as ta

from .protocols import LoggerLike


##


def exception_logging(log, exc_cls=Exception):  # noqa
    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except exc_cls as e:  # noqa
                log.exception('Error in %r', fn)
                raise

        return inner

    return outer


def async_exception_logging(alog, exc_cls=Exception):  # noqa
    def outer(fn):
        @functools.wraps(fn)
        async def inner(*args, **kwargs):
            try:
                return await fn(*args, **kwargs)
            except exc_cls as e:  # noqa
                await alog.exception('Error in %r', fn)
                raise

        return inner

    return outer


##


class LogTimingContext:
    DEFAULT_LOG: ta.ClassVar[ta.Optional[LoggerLike]] = None

    class _NOT_SPECIFIED:  # noqa
        def __new__(cls, *args, **kwargs):  # noqa
            raise TypeError

    def __init__(
            self,
            description: str,
            *,
            log: ta.Union[LoggerLike, ta.Type[_NOT_SPECIFIED], None] = _NOT_SPECIFIED,  # noqa
            level: int = logging.DEBUG,
    ) -> None:
        super().__init__()

        self._description = description
        if log is self._NOT_SPECIFIED:
            log = self.DEFAULT_LOG  # noqa
        self._log: ta.Optional[LoggerLike] = log  # type: ignore
        self._level = level

    def set_description(self, description: str) -> 'LogTimingContext':
        self._description = description
        return self

    _begin_time: float
    _end_time: float

    def __enter__(self) -> 'LogTimingContext':
        self._begin_time = time.time()

        if self._log is not None:
            self._log.log(self._level, f'Begin : {self._description}')  # noqa

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._end_time = time.time()

        if self._log is not None:
            self._log.log(
                self._level,
                f'End : {self._description} - {self._end_time - self._begin_time:0.2f} s elapsed',
            )


log_timing_context = LogTimingContext



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract


T = ta.TypeVar('T')


##


class LoggerMetricUnit(Abstract):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        try:
            mut = LOGGER_METRIC_UNIT_TYPES
        except NameError:
            pass
        else:
            bcs = [bc for bc in mut if issubclass(cls, bc)]
            if len(bcs) != 1:
                raise TypeError(f'{cls.__name__} must be a subclass of exactly one of {mut}, got {bcs}.')

        try:
            mtc = LoggerMetric
        except NameError:
            pass
        else:
            if issubclass(cls, mtc):
                mp = cls.__mro__.index(mtc)
                mup = cls.__mro__.index(LoggerMetricUnit)
                if mup > mp:
                    raise TypeError(f'{cls.__name__} must have Metric before MetricUnit in its MRO.')


class CountLoggerMetricUnit(LoggerMetricUnit):
    @classmethod
    def default_value(cls) -> ta.Optional[float]:
        return 1


class RatioLoggerMetricUnit(LoggerMetricUnit):
    pass


class SecondsLoggerMetricUnit(LoggerMetricUnit):
    pass


class BytesLoggerMetricUnit(LoggerMetricUnit):
    pass


LOGGER_METRIC_UNIT_TYPES: ta.Tuple[ta.Type[LoggerMetricUnit], ...] = (
    CountLoggerMetricUnit,
    RatioLoggerMetricUnit,
    SecondsLoggerMetricUnit,
    BytesLoggerMetricUnit,
)


##


class LoggerMetricTag(Abstract):
    pass


##


class LoggerMetric(Abstract):
    @ta.final
    def __init__(self, value: ta.Optional[float] = None, *tags: LoggerMetricTag) -> None:
        if value is None:
            value = self.default_value()
        if value is None:
            raise ValueError(f'{type(self).__name__} has no default value.')

        self.__value = value
        self.__tags = tags

    @property
    def value(self) -> float:
        return self.__value

    @classmethod
    def default_value(cls) -> ta.Optional[float]:
        return None

    @property
    def tags(self) -> ta.Sequence[LoggerMetricTag]:
        return self.__tags

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.value!r}, {", ".join(map(repr, self.tags))})'

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        try:
            mtt = LOGGER_METRIC_TYPES
        except NameError:
            pass
        else:
            bcs = [bc for bc in mtt if issubclass(cls, bc)]
            if Abstract in cls.__bases__:
                if len(bcs) > 1:
                    raise TypeError(f'{cls.__name__} must be a subclass of at most one of {mtt}, got {bcs}.')
            else:
                if len(bcs) != 1:
                    raise TypeError(f'{cls.__name__} must be a subclass of exactly one of {mtt}, got {bcs}.')

        # if Abstract not in cls.__bases__ and not issubclass(cls, LoggerMetricUnit):
        #     raise TypeError(f'{cls.__name__} must be a subclass of LoggerMetricUnit.')


class CounterLoggerMetric(CountLoggerMetricUnit, LoggerMetric, Abstract):
    pass


class GaugeLoggerMetric(LoggerMetric, Abstract):
    pass


class HistogramLoggerMetric(LoggerMetric, Abstract):
    pass


LOGGER_METRIC_TYPES: ta.Tuple[ta.Type[LoggerMetric], ...] = (
    CounterLoggerMetric,
    GaugeLoggerMetric,
    HistogramLoggerMetric,
)


##


class AnyLoggerMetricCollector(Abstract, ta.Generic[T]):
    @ta.final
    def metric(self, m: LoggerMetric) -> T:
        return self._metric(m)

    @abc.abstractmethod
    def _metric(self, m: LoggerMetric) -> T:
        raise NotImplementedError


class LoggerMetricCollector(AnyLoggerMetricCollector[None], Abstract):
    @abc.abstractmethod
    def _metric(self, m: LoggerMetric) -> None:
        raise NotImplementedError


class AsyncLoggerMetricCollector(AnyLoggerMetricCollector[ta.Awaitable[None]], Abstract):
    @abc.abstractmethod
    def _metric(self, m: LoggerMetric) -> ta.Awaitable[None]:
        raise NotImplementedError


##


class AnyNopLoggerMetricCollector(AnyLoggerMetricCollector[T], Abstract):
    pass


class NopLoggerMetricCollector(AnyNopLoggerMetricCollector[None], LoggerMetricCollector):
    @ta.final
    def _metric(self, m: LoggerMetric) -> None:
        pass


class AsyncNopLoggerMetricCollector(AnyNopLoggerMetricCollector[ta.Awaitable[None]], AsyncLoggerMetricCollector):
    @ta.final
    async def _metric(self, m: LoggerMetric) -> None:
        pass



import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import marshal as msh
from .. import reflect as rfl
from .collection import TypedValues
from .reflect import reflect_typed_values_impls
from .values import ScalarTypedValue
from .values import TypedValue


##


def _build_typed_value_poly(rty: rfl.Type) -> msh.Polymorphism:
    ty: type[TypedValue] = check.issubclass(check.isinstance(rty, type), TypedValue)  # noqa
    check.state(lang.is_abstract_class(ty))
    return msh.polymorphism_from_subclasses(
        ty,
        naming=msh.Naming.SNAKE,
        strip_suffix=msh.AutoStripSuffix,
    )


class TypedValueMarshalerFactory(msh.MarshalerFactoryMethodClass):
    @msh.MarshalerFactoryMethodClass.make_marshaler.register
    def _make_scalar(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if not (
            isinstance(rty, type) and
            issubclass(rty, ScalarTypedValue) and
            not lang.is_abstract_class(rty)
        ):
            return None

        def inner() -> msh.Marshaler:
            dc_rfl = dc.reflect(check.isinstance(rty, type))
            v_rty = check.single(dc_rfl.fields_inspection.generic_replaced_field_annotations.values())
            v_m = ctx.make_marshaler(v_rty)
            return msh.WrappedMarshaler(lambda _, o: o.v, v_m)

        return inner

    @msh.MarshalerFactoryMethodClass.make_marshaler.register
    def _make_abstract(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if not (
            isinstance(rty, type) and
            issubclass(rty, TypedValue) and
            lang.is_abstract_class(rty)
        ):
            return None

        return lambda: msh.make_polymorphism_marshaler(
            _build_typed_value_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


class TypedValueUnmarshalerFactory(msh.UnmarshalerFactoryMethodClass):
    @msh.UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_scalar(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:
        if not (
            isinstance(rty, type) and
            issubclass(rty, ScalarTypedValue) and
            not lang.is_abstract_class(rty)
        ):
            return None

        def inner() -> msh.Unmarshaler:
            dc_rfl = dc.reflect(rty)
            v_rty = check.single(dc_rfl.fields_inspection.generic_replaced_field_annotations.values())
            v_u = ctx.make_unmarshaler(v_rty)
            return msh.WrappedUnmarshaler(lambda _, v: rty(v), v_u)

        return inner

    @msh.UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_abstract(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:  # noqa
        if not (
            isinstance(rty, type) and
            issubclass(rty, TypedValue) and
            lang.is_abstract_class(rty)
        ):
            return None

        return lambda: msh.make_polymorphism_unmarshaler(
            _build_typed_value_poly(rty).impls,
            msh.WrapperTypeTagging(),
            ctx,
        )


##


def _build_typed_values_impls(rty: rfl.Type) -> msh.Impls:
    gty = check.isinstance(rty, rfl.Generic)
    check.is_(gty.cls, TypedValues)

    tv_cls_set = reflect_typed_values_impls(
        check.single(gty.args),
        find_abstract_subclasses=True,
    )

    tv_impls: list[msh.Impl] = [
        msh.Impl(
            tv_cls,
            msh.translate_name(tv_cls.__name__, msh.Naming.SNAKE),
        )
        for tv_cls in tv_cls_set
    ]
    return msh.Impls(tv_impls)


#


def build_typed_values_marshaler(ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> msh.Marshaler:
    tv_m = msh.make_polymorphism_marshaler(
        msh.Impls(_build_typed_values_impls(rty)),
        msh.WrapperTypeTagging(),
        ctx,
    )
    return msh.IterableMarshaler(tv_m)


class TypedValuesMarshalerFactory(msh.MarshalerFactoryMethodClass):
    @msh.MarshalerFactoryMethodClass.make_marshaler.register
    def _make_generic(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if not (isinstance(rty, rfl.Generic) and rty.cls is TypedValues):
            return None
        return lambda: build_typed_values_marshaler(ctx, rty)

    @msh.MarshalerFactoryMethodClass.make_marshaler.register
    def _make_concrete(self, ctx: msh.MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Marshaler] | None:
        if rty is not TypedValues:
            return None
        return lambda: lang.raise_(NotImplementedError)


#


def build_typed_values_unmarshaler(ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> msh.Unmarshaler:
    tv_u = msh.make_polymorphism_unmarshaler(
        msh.Impls(_build_typed_values_impls(rty)),
        msh.WrapperTypeTagging(),
        ctx,
    )
    return msh.IterableUnmarshaler(lambda it: TypedValues(*it), tv_u)  # noqa


class TypedValuesUnmarshalerFactory(msh.UnmarshalerFactoryMethodClass):
    @msh.UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _build(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:
        if not (isinstance(rty, rfl.Generic) and rty.cls is TypedValues):
            return None
        return lambda: build_typed_values_unmarshaler(ctx, rty)

    @msh.UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _build_concrete(self, ctx: msh.UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], msh.Unmarshaler] | None:  # noqa
        if rty is not TypedValues:
            return None
        return lambda: lang.raise_(NotImplementedError)


##


@lang.static_init
def _install_standard_marshaling() -> None:
    msh.install_standard_factories(
        TypedValueMarshalerFactory(),
        TypedValueUnmarshalerFactory(),

        TypedValuesMarshalerFactory(),
        TypedValuesUnmarshalerFactory(),
    )



"""
TODO:
 - max buf size
 - max recursion depth
 - mark start pos of tokens, currently returning end
"""
import dataclasses as dc
import io
import json
import re
import typing as ta

from .... import check
from .... import lang
from ....funcs.genmachine import GenMachine
from .errors import JsonStreamError


if ta.TYPE_CHECKING:
    import unicodedata
else:
    unicodedata = lang.proxy_import('unicodedata')


##


IdentTokenKind: ta.TypeAlias = ta.Literal['IDENT']

ValueTokenKind: ta.TypeAlias = ta.Literal[
    'STRING',
    'NUMBER',
]

VALUE_TOKEN_KINDS = frozenset(check.isinstance(a, str) for a in ta.get_args(ValueTokenKind))

ControlTokenKind: ta.TypeAlias = ta.Literal[
    'LBRACE',
    'RBRACE',
    'LBRACKET',
    'RBRACKET',
    'COMMA',
    'COLON',
]

SpaceTokenKind: ta.TypeAlias = ta.Literal['SPACE']

CommentTokenKind: ta.TypeAlias = ta.Literal['COMMENT']

TokenKind: ta.TypeAlias = ta.Union[  # noqa
    IdentTokenKind,
    ValueTokenKind,
    ControlTokenKind,
    SpaceTokenKind,
    CommentTokenKind,
]


#

ScalarValue: ta.TypeAlias = str | float | int | None

SCALAR_VALUE_TYPES: tuple[type, ...] = tuple(
    check.isinstance(e, type) if e is not None else type(None)
    for e in ta.get_args(ScalarValue)
)


##


class Position(ta.NamedTuple):
    ofs: int
    line: int
    col: int


class Token(ta.NamedTuple):
    kind: TokenKind
    value: ScalarValue
    raw: str | None

    pos: Position

    def __iter__(self):
        raise TypeError


NUMBER_PAT = re.compile(r'-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?')

CONTROL_TOKENS: ta.Mapping[str, TokenKind] = {
    '{': 'LBRACE',
    '}': 'RBRACE',
    '[': 'LBRACKET',
    ']': 'RBRACKET',
    ',': 'COMMA',
    ':': 'COLON',
}

CONST_IDENT_VALUES: ta.Mapping[str, str | float | None] = {
    'NaN': float('nan'),
    '-NaN': float('-nan'),  # distinguished in parsing even if indistinguishable in value
    'Infinity': float('inf'),
    '-Infinity': float('-inf'),

    'true': True,
    'false': False,
    'null': None,
}

MAX_CONST_IDENT_LEN = max(map(len, CONST_IDENT_VALUES))


##


EXPANDED_SPACE_CHARS = (
    '\u0009'
    '\u000A'
    '\u000B'
    '\u000C'
    '\u000D'
    '\u0020'
    '\u00A0'
    '\u2028'
    '\u2029'
    '\uFEFF'
    '\u1680'
    '\u2000'
    '\u2001'
    '\u2002'
    '\u2003'
    '\u2004'
    '\u2005'
    '\u2006'
    '\u2007'
    '\u2008'
    '\u2009'
    '\u200A'
    '\u202F'
    '\u205F'
    '\u3000'
)


##


@dc.dataclass()
class JsonStreamLexError(JsonStreamError):
    message: str

    pos: Position


class JsonStreamLexer(GenMachine[str, Token]):
    def __init__(
            self,
            *,
            include_raw: bool = False,

            allow_extended_space: bool = False,
            include_space: bool = False,

            allow_comments: bool = False,
            include_comments: bool = False,

            allow_single_quotes: bool = False,
            string_literal_parser: ta.Callable[[str], str] | None = None,

            allow_extended_number_literals: bool = False,
            number_literal_parser: ta.Callable[[str], ta.Any] | None = None,

            allow_extended_idents: bool = False,
    ) -> None:
        self._include_raw = include_raw

        self._allow_extended_space = allow_extended_space
        self._include_space = include_space

        self._allow_comments = allow_comments
        self._include_comments = include_comments

        self._allow_single_quotes = allow_single_quotes
        if string_literal_parser is None:
            string_literal_parser = json.loads  # noqa
        self._string_literal_parser = string_literal_parser

        self._allow_extended_number_literals = allow_extended_number_literals
        self._number_literal_parser = number_literal_parser

        self._allow_extended_idents = allow_extended_idents

        self._char_in_str: str | None = None
        self._char_in_str_len: int = 0
        self._char_in_str_pos: int = 0

        self._ofs = 0
        self._line = 1
        self._col = 0

        self._buf = io.StringIO()

        super().__init__(self._do_main())

    @property
    def pos(self) -> Position:
        return Position(
            self._ofs,
            self._line,
            self._col,
        )

    def _advance_pos(self, c: str) -> str:
        if c and len(c) != 1:
            raise JsonStreamError(c)

        self._ofs += 1

        if c == '\n':
            self._line += 1
            self._col = 0
        else:
            self._col += 1

        return c

    def _yield_char_in(self, c: str) -> str:
        if self._char_in_str is not None:
            raise JsonStreamError

        if (cl := len(c)) > 1:
            self._char_in_str = c
            self._char_in_str_len = cl
            self._char_in_str_pos = 1
            c = c[0]

        self._advance_pos(c)

        return c

    def _str_char_in(self) -> str | None:
        if (s := self._char_in_str) is None:
            return None

        if (p := self._char_in_str_pos) >= self._char_in_str_len:
            self._char_in_str = None
            return None

        c = s[p]
        self._char_in_str_pos += 1
        return self._advance_pos(c)

    def _make_tok(
            self,
            kind: TokenKind,
            value: ScalarValue,
            raw: str | None,
            pos: Position,
    ) -> ta.Sequence[Token]:
        tok = Token(
            kind,
            value,
            raw if self._include_raw else None,
            pos,
        )
        return (tok,)

    def _flip_buf(self) -> str:
        raw = self._buf.getvalue()
        self._buf.seek(0)
        self._buf.truncate()
        return raw

    def _raise(self, msg: str, src: Exception | None = None) -> ta.NoReturn:
        raise JsonStreamLexError(msg, self.pos) from src

    def _do_main(self, peek: str | None = None):
        while True:
            if peek is not None:
                c = peek
                peek = None
            else:
                if (c := self._str_char_in()) is None:  # type: ignore[assignment]
                    c = self._yield_char_in((yield None))  # noqa

            if not c:
                return None

            if c.isspace() or (self._allow_extended_space and c in EXPANDED_SPACE_CHARS):
                if self._include_space:
                    yield self._make_tok('SPACE', c, c, self.pos)
                continue

            if c in CONTROL_TOKENS:
                yield self._make_tok(CONTROL_TOKENS[c], c, c, self.pos)
                continue

            if c == '"' or (self._allow_single_quotes and c == "'"):
                return self._do_string(c)

            if c.isdigit() or c == '-' or (self._allow_extended_number_literals and c in '.+'):
                return self._do_number(c)

            if self._allow_comments and c == '/':
                return self._do_comment()

            if self._allow_extended_idents:
                return self._do_extended_ident(c)

            if c in 'tfnIN':
                return self._do_const(c)

            self._raise(f'Unexpected character: {c}')

    def _do_string(self, q: str):
        check.state(self._buf.tell() == 0)
        self._buf.write(q)

        pos = self.pos

        #

        buf = self._buf

        char_in_str = self._char_in_str
        char_in_str_len = self._char_in_str_len
        char_in_str_pos = self._char_in_str_pos
        ofs = self._ofs
        line = self._line
        col = self._col

        def restore_state():
            self._char_in_str = char_in_str
            self._char_in_str_len = char_in_str_len
            self._char_in_str_pos = char_in_str_pos
            self._ofs = ofs
            self._line = line
            self._col = col

        while True:
            c: str | None = None

            while True:
                if char_in_str is not None:
                    if char_in_str_pos >= char_in_str_len:
                        char_in_str = None
                        continue

                    skip_to = char_in_str_len
                    if (qp := char_in_str.find(q, char_in_str_pos)) >= 0 and qp < skip_to:
                        skip_to = qp
                    if (sp := char_in_str.find('\\', char_in_str_pos)) >= 0 and sp < skip_to:
                        skip_to = sp

                    if skip_to != char_in_str_pos:
                        ofs += skip_to - char_in_str_pos
                        if (np := char_in_str.rfind('\n', char_in_str_pos, skip_to)) >= 0:
                            line += char_in_str.count('\n', char_in_str_pos, skip_to)
                            col = skip_to - np - 1
                        else:
                            col += skip_to - char_in_str_pos
                        buf.write(char_in_str[char_in_str_pos:skip_to])

                        if skip_to >= char_in_str_len:
                            char_in_str = None
                            continue
                        char_in_str_pos = skip_to

                    c = char_in_str[char_in_str_pos]
                    char_in_str_pos += 1

                if c is None:
                    try:
                        c = (yield None)
                    except GeneratorExit:
                        restore_state()
                        self._raise('Unexpected end of input')

                    if len(c) > 1:
                        char_in_str = c
                        char_in_str_len = len(char_in_str)
                        char_in_str_pos = 0
                        c = None
                        continue

                if c is None:
                    raise JsonStreamError

                if c and len(c) != 1:
                    raise JsonStreamError(c)

                break

            ofs += 1

            if c == '\n':
                line += 1
                col = 0
            else:
                col += 1

            if not c:
                restore_state()
                self._raise(f'Unterminated string literal: {buf.getvalue()}')

            buf.write(c)
            if c == q:
                # Count consecutive backslashes before this quote
                backslash_count = 0
                buf_val = buf.getvalue()
                check_pos = len(buf_val) - 2  # -2 because we just wrote the quote
                while check_pos >= 0 and buf_val[check_pos] == '\\':
                    backslash_count += 1
                    check_pos -= 1

                # Quote is escaped only if preceded by odd number of backslashes
                if backslash_count % 2 == 0:
                    break

        restore_state()

        #

        raw = self._flip_buf()
        try:
            sv = self._string_literal_parser(raw)
        except json.JSONDecodeError as e:
            self._raise(f'Invalid string literal: {raw!r}', e)

        yield self._make_tok('STRING', sv, raw, pos)

        return self._do_main()

    def _do_number(self, c: str):
        check.state(self._buf.tell() == 0)
        self._buf.write(c)

        pos = self.pos

        while True:
            try:
                if (c := self._str_char_in()) is None:  # type: ignore[assignment]
                    c = self._yield_char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if not c:
                break

            if not (c.isdigit() or c in '.eE+-' or (self._allow_extended_number_literals and c in 'xXabcdefABCDEF')):
                break
            self._buf.write(c)

        raw = self._flip_buf()

        #

        if self._allow_extended_number_literals:
            p = 1 if raw[0] in '+-' else 0
            if (len(raw) - p) > 1 and raw[p] == '0' and raw[p + 1].isdigit():
                self._raise('Invalid number literal')

        if raw == '-' or (self._allow_extended_number_literals and raw == '+'):
            for svs in [
                'Infinity',
                *(['NaN'] if self._allow_extended_number_literals else []),
            ]:
                if c != svs[0]:
                    continue

                if not c:
                    self._raise('Unexpected end of input')

                raw += c
                try:
                    for _ in range(len(svs) - 1):
                        if (c := self._str_char_in()) is None:  # type: ignore[assignment]
                            c = self._yield_char_in((yield None))  # noqa
                        if not c:
                            break
                        raw += c
                except GeneratorExit:
                    self._raise('Unexpected end of input')

                if raw[1:] != svs:
                    self._raise(f'Invalid number format: {raw}')

                if raw[0] == '+':
                    raw = raw[1:]

                yield self._make_tok('IDENT', raw, raw, pos)

                return self._do_main()

        #

        nv: ta.Any

        if (np := self._number_literal_parser) is not None:
            nv = np(raw)

        else:
            if not NUMBER_PAT.fullmatch(raw):
                self._raise(f'Invalid number format: {raw}')

            if '.' in raw or 'e' in raw or 'E' in raw:
                nv = float(raw)
            else:
                nv = int(raw)

        yield self._make_tok('NUMBER', nv, raw, pos)

        #

        if not c:
            return None

        return self._do_main(c)

    def _do_const(self, c: str):
        pos = self.pos
        raw = c
        while True:
            try:
                if (c := self._str_char_in()) is None:  # type: ignore[assignment]
                    c = self._yield_char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            raw += c

            if raw in CONST_IDENT_VALUES:
                break

            if len(raw) > MAX_CONST_IDENT_LEN:
                self._raise(f'Invalid literal: {raw}')

        yield self._make_tok('IDENT', raw, raw, pos)

        return self._do_main()

    def _do_unicode_escape(self):
        try:
            if (c := self._str_char_in()) is None:
                c = self._yield_char_in((yield None))  # noqa
        except GeneratorExit:
            self._raise('Unexpected end of input')

        if c != 'u':
            self._raise('Illegal identifier escape')

        ux = []
        for _ in range(4):
            try:
                if (c := self._str_char_in()) is None:
                    c = self._yield_char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if c not in '0123456789abcdefABCDEF':
                self._raise('Illegal identifier escape')

            ux.append(c)

        return chr(int(''.join(ux), 16))

    def _do_extended_ident(self, c: str):
        check.state(self._buf.tell() == 0)

        if c == '\\':
            c = yield from self._do_unicode_escape()

        elif not (c in '$_' or unicodedata.category(c).startswith('L')):
            self._raise('Illegal identifier start')

        self._buf.write(c)

        pos = self.pos

        while True:
            try:
                if (c := self._str_char_in()) is None:  # type: ignore[assignment]
                    c = self._yield_char_in((yield None))  # noqa
            except GeneratorExit:
                self._raise('Unexpected end of input')

            if c == '\\':
                c = yield from self._do_unicode_escape()
                self._buf.write(c)
                continue

            if not c:
                break

            if c not in '$_\u200C\u200D':
                uc = unicodedata.category(c)
                if not (uc.startswith(('L', 'M', 'N')) or uc == 'Pc'):
                    break

            self._buf.write(c)

        raw = self._flip_buf()

        yield self._make_tok('IDENT', raw, raw, pos)

        return self._do_main(c)

    def _do_comment(self):
        check.state(self._buf.tell() == 0)

        pos = self.pos
        try:
            if (oc := self._str_char_in()) is None:
                oc = self._yield_char_in((yield None))  # noqa
        except GeneratorExit:
            self._raise('Unexpected end of input')

        if oc == '/':
            while True:
                try:
                    if (ic := self._str_char_in()) is None:
                        ic = self._yield_char_in((yield None))  # noqa
                except GeneratorExit:
                    self._raise('Unexpected end of input')

                if not ic or ic == '\n':
                    break

                if self._include_comments:
                    self._buf.write(ic)

            if self._include_comments:
                cmt = self._flip_buf()
                raw = f'//{cmt}\n' if self._include_raw else None
                yield self._make_tok('COMMENT', cmt, raw, pos)

            if not ic:
                return

        elif oc == '*':
            lc: str | None = None
            while True:
                try:
                    if (ic := self._str_char_in()) is None:
                        ic = self._yield_char_in((yield None))  # noqa
                except GeneratorExit:
                    self._raise('Unexpected end of input')

                if lc == '*' and ic == '/':
                    break

                if lc is not None and self._include_comments:
                    self._buf.write(lc)
                lc = ic

            if self._include_comments:
                cmt = self._flip_buf()
                raw = f'/*{cmt}*/' if self._include_raw else None
                yield self._make_tok('COMMENT', cmt, raw, pos)

        else:
            self._raise(f'Unexpected character after comment start: {oc}')

        return self._do_main()



import collections
import typing as ta

from .. import lang


T = ta.TypeVar('T')


##


_MISSING = object()


class PeekIterator(ta.Iterator[T]):
    def __init__(self, it: ta.Iterable[T]) -> None:
        super().__init__()

        self._it = iter(it)
        self._pos = -1
        self._next_item: ta.Any = _MISSING

    _item: T

    def __iter__(self) -> ta.Self:
        return self

    @property
    def done(self) -> bool:
        try:
            self.peek()
        except StopIteration:
            return True
        else:
            return False

    def __next__(self) -> T:
        if self._next_item is not _MISSING:
            self._item = ta.cast(T, self._next_item)
            self._next_item = _MISSING
        else:
            self._item = next(self._it)
        self._pos += 1
        return self._item

    def peek(self) -> T:
        if self._next_item is not _MISSING:
            return ta.cast(T, self._next_item)
        self._next_item = next(self._it)
        return self._next_item

    def maybe_peek(self) -> lang.Maybe[T]:
        try:
            v = self.peek()
        except StopIteration:
            return lang.empty()
        else:
            return lang.just(v)

    def next_peek(self) -> T:
        next(self)
        return self.peek()

    #

    def take_while(self, fn: ta.Callable[[T], bool]) -> ta.Iterator[T]:
        while fn(self.peek()):
            yield next(self)

    def skip_while(self, fn: ta.Callable[[T], bool]) -> None:
        while fn(self.peek()):
            next(self)

    def take_until(self, fn: ta.Callable[[T], bool]) -> ta.Iterator[T]:
        return self.take_while(lambda e: not fn(e))

    def skip_until(self, fn: ta.Callable[[T], bool]) -> None:
        self.skip_while(lambda e: not fn(e))

    def take_through(self, pos: int) -> ta.Iterator[T]:
        return self.take_while(lambda _: self._pos < pos)

    def skip_through(self, pos: int) -> None:
        self.skip_while(lambda _: self._pos < pos)

    def take_to(self, pos: int) -> ta.Iterator[T]:
        return self.take_through(pos - 1)

    def skip_to(self, pos: int) -> None:
        self.skip_through(pos - 1)


class ProxyIterator(ta.Iterator[T]):
    def __init__(self, fn: ta.Callable[[], T]) -> None:
        self._fn = fn

    def __iter__(self) -> ta.Self:
        return self

    def __next__(self) -> T:
        return self._fn()


class PrefetchIterator(ta.Iterator[T]):
    def __init__(self, fn: ta.Callable[[], T] | None = None) -> None:
        super().__init__()

        self._fn = fn
        self._deque: collections.deque[T] = collections.deque()

    def __iter__(self) -> ta.Self:
        return self

    def push(self, item) -> None:
        self._deque.append(item)

    def __next__(self) -> T:
        try:
            return self._deque.popleft()
        except IndexError:
            if self._fn is None:
                raise StopIteration from None
        return self._fn()


class RetainIterator(ta.Iterator[T]):
    def __init__(self, fn: ta.Callable[[], T]) -> None:
        super().__init__()

        self._fn = fn
        self._deque: collections.deque[T] = collections.deque()

    def __iter__(self) -> ta.Self:
        return self

    def pop(self) -> None:
        self._deque.popleft()

    def __next__(self) -> T:
        item = self._fn()
        self._deque.append(item)
        return item



# ruff: noqa: UP045
# @omlish-lite
import typing as ta


##


class PushbackReader:
    """
    Wrap a non-seekable stream (e.g. socket.makefile('rb') or BufferedReader)
    and provide:
      - read(n=-1)
      - readline(limit=-1)
      - readuntil(delim, max_bytes, include_delim=True, chunk_size=8192)

    Extra bytes read past a delimiter are kept in an internal pushback buffer and returned on subsequent reads/readline
    calls.
    """

    def __init__(self, raw):
        self._raw = raw
        self._pushback = bytearray()

    class MaxLengthExceededError(Exception):
        pass

    def read(self, n: int = -1) -> bytes:
        if n == 0:
            return b''

        # read all
        if n < 0:
            out = bytes(self._pushback)
            self._pushback.clear()
            rest = self._raw.read(-1)
            return out + (rest or b'')

        # satisfy from pushback first
        if len(self._pushback) >= n:
            out = bytes(self._pushback[:n])
            del self._pushback[:n]
            return out

        out = bytes(self._pushback)
        self._pushback.clear()
        need = n - len(out)
        chunk = self._raw.read(need) or b''
        return out + chunk

    def readline(self, limit: int = -1) -> bytes:
        # If limit is set, we must not exceed it.
        if limit == 0:
            return b''

        # First, see if pushback already contains a full line.
        nl = self._pushback.find(b'\n')
        if nl != -1:
            end = nl + 1
            if limit > -1:
                end = min(end, limit)
            out = bytes(self._pushback[:end])
            del self._pushback[:end]
            return out

        # Otherwise, accumulate until newline or limit or EOF.
        buf = bytearray()
        if self._pushback:
            buf += self._pushback
            self._pushback.clear()

        while True:
            # Stop if limit reached
            if limit > -1 and len(buf) >= limit:
                return bytes(buf[:limit])

            # Read more
            # Choose a chunk size that respects limit
            to_read = 8192
            if limit > -1:
                to_read = min(to_read, limit - len(buf))
            chunk = self._raw.read1(to_read)
            if not chunk:
                return bytes(buf)

            buf += chunk
            nl = buf.find(b'\n')
            if nl != -1:
                end = nl + 1
                if limit > -1:
                    end = min(end, limit)
                out = bytes(buf[:end])
                # push back anything after the returned part
                if end < len(buf):
                    self._pushback += buf[end:]
                return out

    def readuntil(
            self,
            delim: bytes,
            *,
            max_bytes: ta.Optional[int] = None,
            include_delim: bool = True,
            chunk_size: int = 8192,
    ) -> bytes:
        if not delim:
            raise ValueError('delim must be non-empty')
        if max_bytes is not None and max_bytes < 0:
            raise ValueError('max_bytes must be >= 0')

        buf = bytearray()

        # start with any pushback data
        if self._pushback:
            buf += self._pushback
            self._pushback.clear()

        while True:
            idx = buf.find(delim)
            if idx != -1:
                end = idx + (len(delim) if include_delim else 0)
                out = bytes(buf[:end])
                # push back everything after the delimiter (or after idx+len(delim))
                tail_start = idx + len(delim)
                if tail_start < len(buf):
                    self._pushback += buf[tail_start:]
                return out

            if max_bytes is not None:
                if len(buf) >= max_bytes:
                    raise PushbackReader.MaxLengthExceededError(f'Delimiter not found within {max_bytes} bytes')
                # read another chunk, but don't exceed max_bytes by too much
                to_read = min(chunk_size, max_bytes - len(buf))
            else:
                to_read = chunk_size

            chunk = self._raw.read1(to_read)
            if not chunk:
                raise EOFError('EOF reached before delimiter was found')
            buf += chunk



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta


##


class RawBytesReader(ta.Protocol):
    def read1(self, n: int = -1, /) -> bytes: ...


class BufferedBytesReader(RawBytesReader, ta.Protocol):
    def read(self, n: int = -1, /) -> bytes: ...

    def readall(self) -> bytes: ...


#


class AsyncRawBytesReader(ta.Protocol):
    def read1(self, n: int = -1, /) -> ta.Awaitable[bytes]: ...


class AsyncBufferedBytesReader(AsyncRawBytesReader, ta.Protocol):
    def read(self, n: int = -1, /) -> ta.Awaitable[bytes]: ...

    def readall(self) -> ta.Awaitable[bytes]: ...


##


class FileLikeRawBytesReader:
    """
    Adapter: file-like object -> RawBytesReader-style `read1`.

    This is intentionally permissive: it duck-types common file-like APIs.

    Notes:
      - If the object has `read1`, we use it.
      - Otherwise we fall back to `read`.
      - This is a *raw* reader: it makes no buffering guarantees beyond what the wrapped object provides.
    """

    def __init__(self, f: ta.Any) -> None:
        super().__init__()

        self._f = f

    def read1(self, n: int = -1, /) -> bytes:
        f = self._f
        if hasattr(f, 'read1'):
            return ta.cast(bytes, f.read1(n))
        return ta.cast(bytes, f.read(n))


class FileLikeBufferedBytesReader(FileLikeRawBytesReader):
    """
    Adapter: file-like object -> BufferedBytesReader-style `read/readall`.

    Notes:
      - Uses `readall` if present; otherwise uses `read()` with `-1`.
      - Does not impose additional buffering; it reflects the wrapped object's behavior.
    """

    def read(self, n: int = -1, /) -> bytes:
        return ta.cast(bytes, self._f.read(n))

    def readall(self) -> bytes:
        f = self._f
        if hasattr(f, 'readall'):
            return ta.cast(bytes, f.readall())
        return ta.cast(bytes, f.read())
